/* DFBKana - a kana drill utility for DirectFB
 *
 * Copyright (C) 2001 Andreas Hundt <andi@fischlustig.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
                   
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include <directfb.h>
#include "config.h"

#define DFBCHECK(x...)                                                \
     {                                                                \
          err = x;                                                    \
          if (err != DFB_OK) {                                        \
               fprintf( stderr, "%s <%d>:\n\t", __FILE__, __LINE__ ); \
               DirectFBErrorFatal( #x, err );                         \
          }                                                           \
     }


#define FONT       "decker.ttf"

#define KANAWIDTH  80
#define KANAHEIGHT 50


/* DirectFB interfaces  */
IDirectFB               *dfb;
IDirectFBWindow         *gridwindow;
IDirectFBWindow         *kanawindow;
IDirectFBDisplayLayer   *layer;
IDirectFBInputBuffer    *keybuffer;
IDirectFBImageProvider  *provider;
IDirectFBFont           *font;
IDirectFBFont           *japanese_font = NULL;

/* surface contains all hiragana */
IDirectFBSurface	*hiragana_surface = NULL; 

/* surface contains all katakana */
IDirectFBSurface	*katakana_surface = NULL; 

IDirectFBSurface        *bgsurface;


int xres,yres,fontheight;

#ifdef DEFAULT_JAPANESE_FONT
char *japanese_fontname = JAPANESE_FONT;
#else
char *japanese_fontname = NULL;
#endif

typedef struct _Kana {
     IDirectFBSurface   *hiragana_surface;         /* this is actually a SubSurface */
     IDirectFBSurface   *katakana_surface;         /* this is actually a SubSurface */
     char               *romanji;
     char               *hiragana_char;
     char               *katakana_char;
} Kana;

Kana kanagrid [5][10];

char *romanjitab[50] = 
{
     "a"    , "i"    , "u"     , "e"    , "o"    ,     
     "ka"   , "ki"   , "ku"    , "ke"   , "ko"   ,
     "sa"   , "shi"  , "su"    , "se"   , "so"   ,
     "ta"   , "chi"  , "tsu"   , "te"   , "to"   ,
     "na"   , "ni"   , "nu"    , "ne"   , "no"   ,
     "ha/wa", "hi"   , "hu/fu" , "he/e" , "ho"   ,
     "ma"   , "mi"   , "mu"    , "me"   , "mo"   ,
     "ya"   , "--"   , "yu"    , "--"   , "yo"   ,
     "ra/la", "ri/li", "ru/lu" , "re/le", "ro/lo",
     "wa"   , "--"   , "n/m/ng", "--"   , "wo/o"
};

/* hiragana (UTF-8) */
char *hiraganatab[50] = 
{
    /* a i u e o */
    "\xe3\x81\x82","\xe3\x81\x84","\xe3\x81\x86","\xe3\x81\x88","\xe3\x81\x8a",
    /* ka ki ku ke ko */
    "\xe3\x81\x8b","\xe3\x81\x8d","\xe3\x81\x8f","\xe3\x81\x91","\xe3\x81\x93",
    /* sa shi su se so */
    "\xe3\x81\x95","\xe3\x81\x97","\xe3\x81\x99","\xe3\x81\x9b","\xe3\x81\x9d",
    /* ta chi tsu te to */
    "\xe3\x81\x9f","\xe3\x81\xa1","\xe3\x81\xa4","\xe3\x81\xa6","\xe3\x81\xa8",
    /* na ni nu ne no */
    "\xe3\x81\xaa","\xe3\x81\xab","\xe3\x81\xac","\xe3\x81\xad","\xe3\x81\xae",
    /* ha hi hu he ho */
    "\xe3\x81\xaf","\xe3\x81\xb2","\xe3\x81\xb5","\xe3\x81\xb8","\xe3\x81\xbb",
    /* ma mi mu me mo */
    "\xe3\x81\xbe","\xe3\x81\xbf","\xe3\x82\x80","\xe3\x82\x81","\xe3\x82\x82",
    /* ya yu yo */
    "\xe3\x82\x84",     "--"     ,"\xe3\x82\x86",     "--"     ,"\xe3\x82\x88",
    /* ra ri ru re ro */
    "\xe3\x82\x89","\xe3\x82\x8a","\xe3\x82\x8b","\xe3\x82\x8c","\xe3\x82\x8d",
    /* wa n wo */
    "\xe3\x82\x8f",     "--"     ,"\xe3\x82\x93",     "--"     ,"\xe3\x82\x92"
};

char *katakanatab[50] = 
{
    /* a i u e o */
    "\xe3\x82\xa2","\xe3\x82\xa4","\xe3\x82\xa6","\xe3\x82\xa8","\xe3\x82\xaa",    
    /* ka ki ku ke ko */
    "\xe3\x82\xab","\xe3\x82\xad","\xe3\x82\xaf","\xe3\x82\xb1","\xe3\x82\xb3",
    /* sa shi su se so */
    "\xe3\x82\xb5","\xe3\x82\xb7","\xe3\x82\xb9","\xe3\x82\xbb","\xe3\x82\xbd",    
    /* ta chi tsu te to */
    "\xe3\x82\xbf","\xe3\x83\x81","\xe3\x83\x84","\xe3\x83\x86","\xe3\x83\x88",
    /* na ni nu ne no */
    "\xe3\x83\x8a","\xe3\x83\x8b","\xe3\x83\x8c","\xe3\x83\x8d","\xe3\x83\x8e",    
    /* ha hi hu he ho */
    "\xe3\x83\x8f","\xe3\x83\x92","\xe3\x83\x95","\xe3\x83\x98","\xe3\x83\x9b",    
    /* ma mi mu me mo */
    "\xe3\x83\x9e","\xe3\x83\x9f","\xe3\x83\xa0","\xe3\x83\xa1","\xe3\x83\xa2",
    /* ya yu yo */
    "\xe3\x83\xa4",     "--"     ,"\xe3\x83\xa6",     "--"     ,"\xe3\x83\xa8",
    /* ra ri ru re ro */
    "\xe3\x83\xa9","\xe3\x83\xaa","\xe3\x83\xab","\xe3\x83\xac","\xe3\x83\xad",
    /* wa n wo */
    "\xe3\x83\xaf",     "--"     ,"\xe3\x83\xb3",     "--"     ,"\xe3\x83\xb2"
};

int parse_commandline( int argc, char* argv[] )
{
     int i;
     if (argc && argv) {
          for (i = 1; i < argc; i++) {       
               if (strcmp ((argv)[i], "--help") == 0) {
                    printf( "Usage:\n" );
                    printf( "dfbkana --japanese-font=<filename> to enable anti aliased truetype kana\n" );
                    printf( "(hint: kochi-mincho.ttf is a good choice)\n" );
                    printf( "dfbkana --bitmap-font to force use of bitmap font\n" );
                    
                    return 1;
               }

               if (strncmp ((argv)[i], "--", 2) == 0) {
                    int len = strlen( (argv)[i] ) - 2;

                    if (len) {
                         char *name, *value;

                         name = strdup( (argv)[i] + 2 );
                         value = strchr( name, '=' );  
            
                         if (value) 
                              *value++ = '\0';
                              
                         if (strcmp( name, "japanese-font" ) == 0) {
                              japanese_fontname = strdup( value );
                              free( name );
                              return 0;
                         }
                         
                         if (strcmp( name, "bitmap-font" ) == 0) {
                              japanese_fontname = NULL;
                              free( name );
                              return 0;
                         }                              
                         free( name );
                     }
               }
          }
     }
     return 0;
}

void init_resources( int argc, char *argv[] )
{
     DFBResult err;
     DFBSurfaceDescription dsc;

     srand((long)time(0));

     DFBCHECK(DirectFBInit( &argc, &argv ));

     /* create the super interface */
     DFBCHECK(DirectFBCreate( &dfb ));

     /* create an input buffer for key events */
     DFBCHECK(dfb->CreateInputBuffer( dfb, DICAPS_KEYS, &keybuffer ));

     DFBCHECK(dfb->GetDisplayLayer( dfb, DLID_PRIMARY, &layer ));

     /* load font */
     {
          DFBFontDescription desc;

          desc.flags = DFDESC_HEIGHT;
          desc.height = 25;

          DFBCHECK(dfb->CreateFont( dfb, DATADIR FONT, &desc, &font ));
          DFBCHECK(font->GetHeight( font, &fontheight ));
     }

     /* load japanese font */
     if (japanese_fontname)
     {
          DFBFontDescription desc;

          desc.flags = DFDESC_HEIGHT;
          desc.height = KANAHEIGHT;

          dfb->CreateFont( dfb, japanese_fontname, &desc, &japanese_font );
     }     
     /* load the surface containting all hiragana */
     DFBCHECK(dfb->CreateImageProvider( dfb, DATADIR"hiragana.png",
                                        &provider ));

     DFBCHECK (provider->GetSurfaceDescription (provider, &dsc));
     DFBCHECK(dfb->CreateSurface( dfb, &dsc, &hiragana_surface ));

     DFBCHECK(provider->RenderTo( provider, hiragana_surface ));
     provider->Release( provider ); 

 
     /* load the surface containting all katakana*/
     DFBCHECK(dfb->CreateImageProvider( dfb, DATADIR"katakana.png",
                                        &provider ));

     DFBCHECK (provider->GetSurfaceDescription (provider, &dsc));
     DFBCHECK(dfb->CreateSurface( dfb, &dsc, &katakana_surface ));
     
     DFBCHECK(provider->RenderTo( provider, katakana_surface ));
     provider->Release( provider ); 

     layer->EnableCursor ( layer, 1 );     
}
void set_backgroundimage( char* filename )
{
          DFBResult                err;
          
          IDirectFBImageProvider  *provider;
          IDirectFBSurface        *bgsurface;
          DFBSurfaceDescription    desc;
          DFBDisplayLayerConfig    layer_config;


          layer->GetConfiguration( layer, &layer_config );

          DFBCHECK(dfb->CreateImageProvider( dfb, filename, &provider ));

          desc.flags  = DSDESC_WIDTH | DSDESC_HEIGHT;
          desc.width  = layer_config.width;
          desc.height = layer_config.height;

          DFBCHECK(dfb->CreateSurface( dfb, &desc, &bgsurface ) );
          DFBCHECK(provider->RenderTo( provider, bgsurface ));
          
          layer->SetBackgroundImage( layer, bgsurface );
          layer->SetBackgroundMode( layer, DLBM_IMAGE );
                   
          provider->Release( provider );
}


void init_gridwindow()
{
     DFBResult err;
     
     DFBWindowDescription desc;

     desc.flags = ( DWDESC_POSX | DWDESC_POSY | DWDESC_WIDTH | DWDESC_HEIGHT );
     
     desc.posx   = 275;
     desc.posy   = 30;
     desc.width  = (KANAWIDTH+1)  *  5 + 1;
     desc.height = (KANAHEIGHT+1) * 10 + 1;

     DFBCHECK(layer->CreateWindow( layer, &desc, &gridwindow ) );

     gridwindow->SetOpacity( gridwindow, 0xB0 );
}

void init_kanawindow()
{
     DFBResult err;
     
     DFBWindowDescription desc;

     desc.flags = ( DWDESC_POSX | DWDESC_POSY | DWDESC_WIDTH | DWDESC_HEIGHT );
     
     desc.posx   = 100;
     desc.posy   = 100;
     desc.width  = KANAWIDTH;
     desc.height = KANAHEIGHT;


     DFBCHECK(layer->CreateWindow( layer, &desc, &kanawindow ) );

     kanawindow->SetOpacity( kanawindow, 0xB0 );
}

void init_kanagrid()
{
     int x;
     int y;
     
     int i = 0;
     
     for (y=0; y<10; y++) {
          for (x=0; x<5; x++) {
               DFBRectangle rect = { 0, 0, KANAWIDTH, KANAHEIGHT };
	       
               rect.x = x*KANAWIDTH;
               rect.y = y*KANAHEIGHT;
               hiragana_surface->GetSubSurface( hiragana_surface,
                                                &rect,
                                                &kanagrid[x][y].hiragana_surface);

               katakana_surface->GetSubSurface( katakana_surface,
                                                &rect,
                                                &kanagrid[x][y].katakana_surface);
                                                
               kanagrid[x][y].romanji = romanjitab[i];
               kanagrid[x][y].hiragana_char = hiraganatab[i];
               kanagrid[x][y].katakana_char = katakanatab[i];
               i++;
          }
     }	     
}

void redraw_gridwindow() 
{
     int x;
     int y;
     
     IDirectFBSurface *gridsurface; 
     
     gridwindow->GetSurface( gridwindow, &gridsurface );
     
     gridsurface->SetFont( gridsurface, font );
     gridsurface->SetColor( gridsurface, 0xFF, 0x00, 0x00, 0xFF );
	     
     for (x = 0; x < (KANAWIDTH+1)*6; x+=(KANAWIDTH+1) )
          gridsurface->DrawLine( gridsurface, x, 0, x, (KANAHEIGHT+1)*10 );
     	  
     for (y = 0; y < (KANAHEIGHT+1)*11; y+=(KANAHEIGHT+1) )
          gridsurface->DrawLine( gridsurface, 0 , y, (KANAWIDTH+1)*5, y );

     gridsurface->SetColor( gridsurface, 0xFF, 0xFF, 0xFF, 0xFF );
     /* fill romanji */
     for (y=0; y<10; y++) {
          for (x=0; x<5; x++) {
               gridsurface->DrawString( gridsurface,  kanagrid[x][y].romanji,
                                    -1, x*(KANAWIDTH +1) + KANAWIDTH /2,
                                        y*(KANAHEIGHT+1) + KANAHEIGHT/2
                                                         - fontheight/2,
                                        DSTF_CENTER | DSTF_TOP );
          }
     }

     gridsurface->Flip( gridsurface, NULL, 0 );
     gridsurface->Release( gridsurface );
}

int update_kanawindow( int kana_nr, int kanamode )
{
     int x;
     int y;
     IDirectFBSurface *kanawindow_surface;

     if (kana_nr == -1) {
          do {
               kana_nr = rand()%50;
          } while ( kana_nr==36 || kana_nr==38 || kana_nr==46 || kana_nr==48 );
     }
          
     x = kana_nr%5;
     y = kana_nr/5;
     kanawindow->GetSurface( kanawindow, &kanawindow_surface );
	       
     if (japanese_font) {
          if (kanamode == 0)  {
               kanawindow_surface->SetColor( kanawindow_surface,
                                             0xFF, 0xFF, 0xFF, 0xFF );

               kanawindow_surface->FillRectangle( kanawindow_surface, 
                                                 0, 0, KANAWIDTH, KANAHEIGHT );
                                             
               kanawindow_surface->SetColor( kanawindow_surface,
                                             0x00, 0x00, 0x00, 0xFF );

               kanawindow_surface->SetFont( kanawindow_surface, japanese_font );
               kanawindow_surface->DrawString( kanawindow_surface,
                                               hiraganatab[kana_nr],
                                               -1,
                                               KANAWIDTH/2,
                                               0,
                                               DSTF_CENTER | DSTF_TOP);
          
          }
          else {
               kanawindow_surface->SetColor( kanawindow_surface,
                                             0xFF, 0xFF, 0xFF, 0xFF );

               kanawindow_surface->FillRectangle( kanawindow_surface, 
                                                 0, 0, KANAWIDTH, KANAHEIGHT );
                                             
               kanawindow_surface->SetColor( kanawindow_surface,
                                             0x00, 0x00, 0x00, 0xFF );

               kanawindow_surface->SetFont( kanawindow_surface, japanese_font );
               kanawindow_surface->DrawString( kanawindow_surface,
                                               katakanatab[kana_nr],
                                               -1,
                                               KANAWIDTH/2,
                                               0,
                                               DSTF_CENTER | DSTF_TOP);
          }
     }
     else {
          if (kanamode == 0) {
               kanawindow_surface->Blit( kanawindow_surface, 
                                         kanagrid[x][y].hiragana_surface,
                                         NULL, 0, 0 );
          }
          else {
               kanawindow_surface->Blit( kanawindow_surface, 
                                         kanagrid[x][y].katakana_surface,
                                         NULL, 0, 0 );

          }
     }
     
     kanawindow_surface->Flip( kanawindow_surface, NULL, 0 );
     return ((y * 5) + x );
     
     kanawindow_surface->Release( kanawindow_surface );
}

void highlight( int kana_nr, int color )
{
     IDirectFBSurface *gridsurface;     
     gridwindow->GetSurface( gridwindow, &gridsurface );
     
     if (color)
          gridsurface->SetColor( gridsurface, 0xFF, 0x00, 0x00, 0x20 );
     else
          gridsurface->SetColor( gridsurface, 0x00, 0xFF, 0x00, 0x20 );
     
     gridsurface->SetDrawingFlags( gridsurface, DSDRAW_BLEND );
 
     gridsurface->FillRectangle( gridsurface, 1+ kana_nr%5 * (KANAWIDTH+1), 
                                              1+ kana_nr/5 * (KANAHEIGHT+1),                                              
					      KANAWIDTH,
					      KANAHEIGHT );
     
     gridsurface->SetDrawingFlags( gridsurface, DSDRAW_NOFX );

     gridsurface->SetFont( gridsurface, font );

     if (color)
          gridsurface->SetColor( gridsurface, 0xFF, 0x00, 0x00, 0xFF );
     else
          gridsurface->SetColor( gridsurface, 0x00, 0xFF, 0x00, 0xFF );
     
     gridsurface->DrawString( gridsurface, romanjitab[kana_nr],
                              -1, (kana_nr%5)*(KANAWIDTH +1) + KANAWIDTH/2,
                                  (kana_nr/5)*(KANAHEIGHT+1) + KANAHEIGHT/2
                                                             - fontheight/2,
                                  DSTF_CENTER | DSTF_TOP);

     gridsurface->Release( gridsurface );
}

void kanawindow_moveto( int kana_nr )
{
     int grid_x = 0;
     int grid_y = 0;

     if ((kana_nr < -1) || (kana_nr > 49))
          return;
          
     if (kana_nr == -1) {
         kanawindow->MoveTo( kanawindow, 100, 100 );
     }
     else {          
          gridwindow->GetPosition( gridwindow, &grid_x, &grid_y );

          kanawindow->MoveTo( kanawindow, grid_x+1 + (kana_nr%5) * (KANAWIDTH+1),
                                        grid_y+1 + (kana_nr/5) * (KANAHEIGHT+1));
     }
}

void solve_kana( int kana_nr, int guessed_kana_nr, int delay )
{
     if (kana_nr == guessed_kana_nr) 
          highlight( kana_nr , 0 );
     else {
          highlight( kana_nr , 1 );
          delay = 1; /* always wait a second if guess was wrong */
     }
       
     kanawindow_moveto( kana_nr );
     sleep( delay );
     kanawindow_moveto( -1 );
}

/*
 * deinitializes resources and DirectFB
 */

void deinit_resources()
{
     keybuffer->Release( keybuffer );
     
     katakana_surface->Release( katakana_surface );
     hiragana_surface->Release( hiragana_surface );
     if (bgsurface)
           bgsurface->Release( bgsurface );     

     kanawindow->Release( kanawindow );
     gridwindow->Release( gridwindow );

     layer->Release( layer );

     dfb->Release( dfb );
}

int main( int argc, char *argv[] )
{
     int quit = 0;
     int kana_nr = 0;
     int kana_mode = 0;
     int kana_window_pos = -1;
     int background_mode = 1;
     
     if (parse_commandline( argc, argv ))
          exit(1);
          
     init_resources( argc, argv );
     
     set_backgroundimage( DATADIR"DoNotEat.jpg" );
     
     init_kanagrid();
     
     init_gridwindow(); 
     init_kanawindow();
     
     redraw_gridwindow();

     kana_nr = update_kanawindow( -1, kana_mode );
     
     while (!quit) {
          DFBWindowEvent evt;
          DFBInputEvent  kevt;
          
	  while (gridwindow->GetEvent( gridwindow, &evt ) == DFB_OK) {
               if (evt.type == DWET_BUTTONDOWN) {
                    int guessed_kana_nr = (evt.y/(KANAHEIGHT+1))*5 
                                         +(evt.x/(KANAWIDTH+1));
                                         
                    solve_kana( kana_nr, guessed_kana_nr, 1 );
                    kana_window_pos = -1;
                    kana_nr = update_kanawindow( -1, kana_mode );
               }
               if (evt.type == DWET_WHEEL) {
                    unsigned char opacity;
                    int tmp_opacity;
                    
                    gridwindow->GetOpacity( gridwindow, &opacity );
                    tmp_opacity = opacity + evt.step * 4;
                    if (tmp_opacity > 255)
                         tmp_opacity = 255;
                    if (tmp_opacity <= 0)
                         tmp_opacity = 1;
                         
                    opacity = tmp_opacity;
                        
                    gridwindow->SetOpacity( gridwindow, opacity );
               }
          }

          while (kanawindow->GetEvent( kanawindow, &evt ) == DFB_OK) {
               if (evt.type == DWET_BUTTONDOWN) {
                    kana_mode= !kana_mode;
                    update_kanawindow( kana_nr, kana_mode );
               }
          }
          
          while (keybuffer->GetEvent( keybuffer, &kevt ) == DFB_OK) {  	
               if (kevt.type == DIET_KEYPRESS)
                    switch (kevt.keycode) {
                    case DIKC_LEFT:
                         if (--kana_window_pos < -1)
                              kana_window_pos = -1;
                         kanawindow_moveto( kana_window_pos );
                         break;
                    case DIKC_RIGHT:
                         if (++kana_window_pos > 49)
                              kana_window_pos = 49;
                         kanawindow_moveto( kana_window_pos );
                         break;
                    case DIKC_UP:
                         kana_window_pos-=5;
                         if (kana_window_pos < -1)
                              kana_window_pos = -1;
                         kanawindow_moveto( kana_window_pos );
                         break;
                    case DIKC_DOWN:
                         kana_window_pos+=5;
                         if (kana_window_pos > 49)
                              kana_window_pos = 49;
                         kanawindow_moveto( kana_window_pos );
                         break;
                    case DIKC_ENTER:
                    case DIKC_OK:
                         if (kana_window_pos != -1) {
                              solve_kana( kana_nr, kana_window_pos, 0 );
                              kana_window_pos = -1;
                              kana_nr = update_kanawindow( -1, kana_mode );
                         }     
                         break;
                    case DIKC_Q:
                         kana_mode = !kana_mode;
                         update_kanawindow( kana_nr, kana_mode );
                         break;
                    case DIKC_1:
                         kana_mode = 0;
                         update_kanawindow( kana_nr, kana_mode );
                         break;
                    case DIKC_2:
                         kana_mode = 1;
                         update_kanawindow( kana_nr, kana_mode );
                         break;
                    case DIKC_B:
                         background_mode = !background_mode;
                         if (background_mode)
                              layer->SetBackgroundMode( layer, DLBM_IMAGE );
                         else
                              layer->SetBackgroundMode( layer, DLBM_COLOR );                                                                                    
                         break;
                    case DIKC_ESCAPE:
                         quit =1;
                    default:
                         break;
               }
          }
          usleep( 20000 );
     }
		     
     deinit_resources();
     return 42;
}
